from os.path import expanduser
import json

home = expanduser("~")


class UserConfig(object):
    def __init__(self, **kwargs):
        super(UserConfig, self).__init__()
        self.manager_port = kwargs.get("quantumMachinesManager_port", 9510)
        self.manager_host = kwargs.get("quantumMachinesManager_host", "localhost")
        self.strict_healthcheck = kwargs.get(
            "quantumMachinesManager_strict_healthcheck", True
        )
        self.user_token = kwargs.get("quantumMachinesManager_user_token", "")


def _load_config_file():
    try:
        with open(home + "/.qm/config.json", "r+") as file:
            return json.load(file)
    except Exception:
        return {}


def load_user_config():
    thejson = _load_config_file()
    return UserConfig(**thejson)



import asyncio
import threading


def run_async(coro):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop and loop.is_running():
        thread = RunThread(coro)
        thread.start()
        thread.join()
        return thread.result
    else:
        return asyncio.run(coro)


class RunThread(threading.Thread):
    def __init__(self, func):
        self.func = func
        self.result = None
        super().__init__()

    def run(self):
        self.result = asyncio.run(self.func)



from qm.QuaNodeVisitor import QuaNodeVisitor


class ExpressionSerializingVisitor(QuaNodeVisitor):
    def __init__(self) -> None:
        self._out = ""
        super().__init__()

    def _default_visit(self, node):
        print("missing expression: " + node.DESCRIPTOR.full_name)
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_VarRefExpression(self, node):
        self._out = node.name if node.name else f"IO{node.ioNumber}"

    def visit_qm_grpc_qua_QuaProgram_ArrayVarRefExpression(self, node):
        self._out = node.name

    def visit_qm_grpc_qua_QuaProgram_ArrayCellRefExpression(self, node):
        var = ExpressionSerializingVisitor.serialize(node.arrayVar)
        index = ExpressionSerializingVisitor.serialize(node.index)
        self._out = f"{var}[{index}]"

    def visit_qm_grpc_qua_QuaProgram_LiteralExpression(self, node):
        self._out = node.value

    def visit_qm_grpc_qua_QuaProgram_AssignmentStatement_Target(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_AnalogProcessTarget_ScalarProcessTarget(
        self, node
    ):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_AnyScalarExpression(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_SaveStatement_Source(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_BinaryExpression(self, node):
        left = ExpressionSerializingVisitor.serialize(node.left)
        right = ExpressionSerializingVisitor.serialize(node.right)
        sop = node.op
        if sop == node.ADD:
            op = "+"
        elif sop == node.SUB:
            op = "-"
        elif sop == node.GT:
            op = ">"
        elif sop == node.LT:
            op = "<"
        elif sop == node.LET:
            op = "<="
        elif sop == node.GET:
            op = ">="
        elif sop == node.EQ:
            op = "=="
        elif sop == node.MULT:
            op = "*"
        elif sop == node.DIV:
            op = "/"
        elif sop == node.OR:
            op = "|"
        elif sop == node.AND:
            op = "&"
        elif sop == node.XOR:
            op = "^"
        elif sop == node.SHL:
            op = "<<"
        elif sop == node.SHR:
            op = ">>"
        else:
            raise Exception("Unsupported operator " + sop)
        self._out = f"{left}{op}{right}"

    @staticmethod
    def serialize(node) -> str:
        visitor = ExpressionSerializingVisitor()
        visitor.visit(node)
        return visitor._out



import json
import grpc

from qm.QmJob import QmJob
from qm._errors import FailedToExecuteJobException, ConfigValidationException
from qm.server_detector import detect_server
from qm.utils import _level_map, _set_compiler_options
from qm import _Program
from qm.logger import logger

from qm.QuantumMachine import QuantumMachine
from qm.pb import frontend_pb2_grpc
from google.protobuf import empty_pb2

from qm.pb.frontend_pb2 import (
    ResetDataProcessingRequest,
    SimulationRequest,
    PerformHalDebugCommandRequest,
)
from qm.pb.qm_manager_pb2 import OpenQuantumMachineRequest, GetQuantumMachineRequest
from qm._controller import Controller
from qm.persistence import SimpleFileStore, BaseStore
from qm.program import load_config
from qm._user_config import load_user_config
from qm.program.ConfigBuilder import convert_msg_to_config
from qm.simulate.interface import (
    SimulationConfig,
    InterOpxAddress,
    InterOpxChannel,
)


class QuantumMachinesManager(object):
    def __init__(self, host=None, port=None, **kargs):
        """
        :param string host: Host where to find the QM orchestrator. If ``None``, local settings are used
        :param port: Port where to find the QM orchestrator. If None, local settings are used
        """
        super(QuantumMachinesManager, self).__init__()
        self._log = logger

        config = load_user_config()
        if host is None:
            host = config.manager_host

        credentials = kargs.get("credentials", None)
        user_token = config.user_token
        server_details, tried_ports = detect_server(
            user_token=user_token,
            credentials=credentials,
            host=host,
            port_from_user_config=config.manager_port,
            user_provided_port=port,
        )

        if server_details is None:
            message = (
                "Failed to connect to QuantumMachines server. Tried connecting to %s"
                % ",".join([f"{host}:{port}" for port in tried_ports])
            )
            self._log.error(message)
            raise Exception(message)

        store: BaseStore = kargs.get("store")
        if not isinstance(store, BaseStore):
            root = kargs.get("file_store_root", ".")
            store = SimpleFileStore(root)
        self._store = store

        self._server_details = server_details
        self._caps = server_details.capabilities
        self._channel = intercept_channel = server_details.grpc_channel
        self._frontend = frontend_pb2_grpc.FrontendStub(intercept_channel)

        raise_on_error = config.strict_healthcheck is not False
        if "log_level" in kargs:
            new_level = kargs.get("log_level")
            try:
                logger.setLevel(new_level)
            except ValueError:
                logger.warning(
                    "Failed to set log level. level '%s' is not recognized", new_level
                )
        self.perform_healthcheck(raise_on_error)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._close()
        return False

    @property
    def store(self) -> BaseStore:
        return self._store

    def perform_healthcheck(self, strict=True):
        """
        Perform a health check against the QM server.

        :param strict: Will raise an exception if health check failed
        """
        with _grpc_context():
            self._log.info("Performing health check")
            req = empty_pb2.Empty()
            res = self._frontend.HealthCheck(req)

            for msg in res.warningMessages:
                self._log.warning("  HC Warning: " + msg)

            if res.ok:
                self._log.info("Health check passed")
            else:
                self._log.error("Health check FAILED!")
                for msg in res.message:
                    self._log.error("  HC Error: " + msg)

                for msg in res.errorMessages:
                    self._log.error("  HC Error: " + msg)
                if strict:
                    raise Exception("Health check failed")

    def version(self):
        """
        :return: The QM server version
        """
        server_version = self._server_details.qop_version
        from qm.version import __version__

        return {"client": __version__, "server": server_version}

    def reset_data_processing(self):
        """
        Stops current data processing for ALL running jobs

        :return:

        """
        with _grpc_context():
            req = ResetDataProcessingRequest()
            self._frontend.ResetDataProcessing(req)

    def close(self):
        self._close()

    def _close(self):
        self._channel.close()

    def open_qm(self, config, close_other_machines=True):
        """
        Opens a new quantum machine. A quantum machine can use multiple OPXes, and a
        single OPX can also be used by multiple quantum machines as long as they do not
        share the same physical resources (input/output ports) as defined in the config.

        :param config: The config that will be used by the quantum machine
        :param close_other_machines: When set to to true (default) any open quantum
                            machines will be closed. This simplify the workflow,
                            but does not enable opening more than one quantum machine.
        :return: A quantum machine obj that can be used to execute programs
        """

        return self._open_qm_pb(load_config(config), close_other_machines)

    @staticmethod
    def _validate_config(
        pb_config,
        *,
        supports_multiple_inputs_for_element: bool,
        supports_analog_delay: bool,
    ):

        if not supports_multiple_inputs_for_element:
            for el_name, el in list(pb_config.v1beta.elements.items()):
                if el.HasField("multipleInputs"):
                    raise ConfigValidationException(
                        f"Server does not support multiple inputs for elements used in '{el_name}'"
                    )

        if not supports_analog_delay:
            for con_name, con in list(pb_config.v1beta.controllers.items()):
                for port_id, port in list(con.analogOutputs.items()):
                    if port.delay != 0:
                        raise ConfigValidationException(
                            f"Server does not support analog delay used in controller '{con_name}', port {port_id}"
                        )

    def _open_qm_pb(self, pb_config, close_other_machines=False):
        QuantumMachinesManager._validate_config(
            pb_config,
            supports_multiple_inputs_for_element=self._caps.supports_multiple_inputs_for_element,
            supports_analog_delay=self._caps.supports_analog_delay,
        )
        with _grpc_context():
            request = OpenQuantumMachineRequest()

            if close_other_machines:
                request.always = True
            else:
                request.never = True

            request.config.CopyFrom(pb_config)

            result = self._frontend.OpenQuantumMachine(request)
            if not result.success:
                exception = Exception("Can not open QM. Please see previous errors")
                for err in result.configValidationErrors:
                    self._log.error(
                        'CONFIG ERROR in key "%s" [%s] : %s',
                        err.path,
                        err.group,
                        err.message,
                    )

                for err in result.physicalValidationErrors:
                    self._log.error(
                        'PHYSICAL CONFIG ERROR in key "%s" [%s] : %s',
                        err.path,
                        err.group,
                        err.message,
                    )

                exception.errors = [
                    (it.group, it.path, it.message)
                    for it in result.configValidationErrors
                ]
                exception.errors += [
                    (it.group, it.path, it.message)
                    for it in result.physicalValidationErrors
                ]

                raise exception

            parsed_config = convert_msg_to_config(request.config)
            return QuantumMachine(result.machineID, request.config, parsed_config, self)

    def open_qm_from_file(self, filename, close_other_machines=True):
        """
        Opens a new quantum machine with config taken from a file on the local file system

        :param filename: The path to the file that contains the config
        :param close_other_machines: Flag whether to close all other running machines
        :return: A quantum machine obj that can be used to execute programs
        """
        with open(filename) as json_file:
            json1_str = json_file.read()

            def remove_nulls(d):
                return {k: v for k, v in d.items() if v is not None}

            config = json.loads(json1_str, object_hook=remove_nulls)
        return self.open_qm(config, close_other_machines)

    def simulate(self, config, program, simulate, **kwargs):
        """
        :param config: A QM config
        :param program: A QUA ``program()`` object to execute
        :param simulate: A ``SimulationConfig`` configuration object
        :param kwargs: additional parameters to pass to execute
        :return: a ``QmJob`` object (see QM Job API).
        """

        if type(program) is not _Program:
            raise Exception("program argument must be of type qm.program.Program")

        with _grpc_context():
            request = SimulationRequest()
            msg_config = load_config(config)
            request.config.CopyFrom(msg_config)

            if type(simulate) is SimulationConfig:
                request.simulate.SetInParent()
                request.simulate.duration = simulate.duration
                request.simulate.includeAnalogWaveforms = (
                    simulate.include_analog_waveforms
                )
                request.simulate.includeDigitalWaveforms = (
                    simulate.include_digital_waveforms
                )
                request.simulate.extraProcessingTimeoutMs = (
                    simulate.extraProcessingTimeoutInMs
                )
                simulate.update_simulate_request(request)

                for connection in simulate.controller_connections:
                    con = request.controllerConnections.add()

                    if type(connection.source) is not type(connection.target):
                        raise Exception(
                            f"Unsupported InterOpx connection. Source is "
                            f"{type(connection.source).__name__} but target is "
                            f"{type(connection.target).__name__}"
                        )

                    if isinstance(connection.source, InterOpxAddress):
                        con.addressToAddress.SetInParent()
                        con.addressToAddress.source.SetInParent()
                        con.addressToAddress.source.controller = (
                            connection.source.controller
                        )
                        con.addressToAddress.source.left = (
                            connection.source.is_left_connection
                        )
                        con.addressToAddress.target.SetInParent()
                        con.addressToAddress.target.controller = (
                            connection.target.controller
                        )
                        con.addressToAddress.target.left = (
                            connection.target.is_left_connection
                        )
                    elif isinstance(connection.source, InterOpxChannel):
                        con.channelToChannel.SetInParent()
                        con.channelToChannel.source.SetInParent()
                        con.channelToChannel.source.controller = (
                            connection.source.controller
                        )
                        con.channelToChannel.source.channelNumber = (
                            connection.source.channel_number
                        )
                        con.channelToChannel.target.SetInParent()
                        con.channelToChannel.target.controller = (
                            connection.target.controller
                        )
                        con.channelToChannel.target.channelNumber = (
                            connection.target.channel_number
                        )
                    else:
                        raise Exception(
                            f"Unsupported InterOpx connection. Source is "
                            f"{type(connection.source).__name__}. Supported types are "
                            f"InterOpxAddress "
                            f"or InterOpxChannel"
                        )

            request.highLevelProgram.CopyFrom(program.build(msg_config))
            _set_compiler_options(request, **kwargs)

            logger.info("Simulating program")

            response = self._frontend.Simulate(request)

            messages = [
                (_level_map[msg.level], msg.message) for msg in response.messages
            ]

            config_messages = [
                (_level_map[msg.level], msg.message)
                for msg in response.configValidationErrors
            ]

            job_id = response.jobId

            for lvl, msg in messages:
                logger.log(lvl, msg)

            for lvl, msg in config_messages:
                logger.log(lvl, msg)

            if not response.success:
                logger.error("Job " + job_id + " failed. Failed to execute program.")
                for error in response.simulated.errors:
                    logger.error(f"Simulation error: {error}")
                raise FailedToExecuteJobException(job_id)

            return QmJob(self, job_id, response)

    def list_open_quantum_machines(self):
        """
        Return a list of open quantum machines. (Returns only the ids, use ``get_qm(...)`` to get the machine object)

        :return: The ids list
        """
        with _grpc_context():
            request = empty_pb2.Empty()
            open_qms = []
            for qm_id in self._frontend.ListOpenQuantumMachines(request).machineIDs:
                open_qms.append(qm_id)
            return open_qms

    def get_qm(self, machine_id):
        """
        Gets an open quantum machine object with the given machine id

        :param machine_id: The id of the open quantum machine to get
        :return: A quantum machine obj that can be used to execute programs
        """
        with _grpc_context():
            request = GetQuantumMachineRequest()
            request.machineID = machine_id
            result = self._frontend.GetQuantumMachine(request)
            if not result.success:
                self._raise_get_qm_errors(result)
            parsed_config = convert_msg_to_config(result.config)
            return QuantumMachine(result.machineID, result.config, parsed_config, self)

    def _get_qm_config(self, machine_id):
        with _grpc_context():
            request = GetQuantumMachineRequest()
            request.machineID = machine_id
            result = self._frontend.GetQuantumMachine(request)
            if not result.success:
                self._raise_get_qm_errors(result)
            return {
                "pb_config": result.config,
                "config": convert_msg_to_config(result.config),
            }

    def _raise_get_qm_errors(self, result):
        error_str = ""
        for err in result.errors:
            error_str = error_str + err.message + "\n"
        exception = Exception(error_str)
        exception.errors = [(it.code, it.message) for it in result.errors]
        raise exception

    def close_all_quantum_machines(self):
        """
        Closes ALL open quantum machines
        """
        with _grpc_context():
            request = empty_pb2.Empty()
            result = self._frontend.CloseAllQuantumMachines(request)
            if not result.success:
                exception = Exception(
                    "Can not close all quantum machines. Please see previous errors"
                )
                for err in result.errors:
                    self._log.error(err.message)

                exception.errors = [(it.code, it.message) for it in result.errors]
                raise exception

    def get_controllers(self):
        """
        Returns a list of all the controllers that are available
        """
        with _grpc_context():
            request = empty_pb2.Empty()
            result = self._frontend.GetControllers(request)
            controllers_list = []
            for controller in result.controllers:
                controllers_list.append(Controller.build_from_message(controller))
            return controllers_list

    def validate_version(self, host):
        # SDK version is not coupled to a specific version anymore.
        # We check for capabilities
        pass

    def clear_all_job_results(self):
        """
        Deletes all data from all previous jobs
        """
        with _grpc_context():
            request = empty_pb2.Empty()
            self._frontend.ClearAllJobResults(request)

    def _send_debug_command(self, controller_name, command):
        with _grpc_context():
            request = PerformHalDebugCommandRequest()
            request.controllerName = controller_name
            request.command = command
            response = self._frontend.PerformHalDebugCommand(request)
            if not response.success:
                raise Exception(response.response)
            return response.response


class _grpc_context(object):
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None and issubclass(exc_type, grpc.RpcError):
            # raise RuntimeError("Failed to contact QM manager")
            try:
                print(exc_val)
                details = ": " + exc_val.details()
            except RuntimeError as e:
                print(e)
                details = ""
            raise RuntimeError("Failed to contact QM manager" + details)
        pass



class QuaNodeVisitor:
    def _default_enter(self, node):
        return True

    def _default_leave(self, node):
        pass

    def _default_visit(self, node):
        fields = node.ListFields()
        for (_descriptor, field) in fields:
            self.__visit(field)

    def __call(self, name, node, t):
        attr_name = f"{t}_{name}".replace(".", "_")
        attr = getattr(self, attr_name, None)
        if attr:
            ret = attr(node)
            if t == "enter":
                return ret
        elif t == "enter":
            return self._default_enter(node)
        elif t == "leave":
            self._default_leave(node)
        elif t == "visit":
            self._default_visit(node)
        else:
            raise Exception(
                f"unknown call type {t}. only 'enter', 'leave' or 'visit' are supported"
            )

    def __call_enter(self, name, node):
        return self.__call(name, node, "enter")

    def __call_visit(self, name, node):
        return self.__call(name, node, "visit")

    def __call_leave(self, name, node):
        return self.__call(name, node, "leave")

    def visit(self, node):
        type_name = type(node).__name__
        type_module = type(node).__module__
        type_fullname = f"{type_module}.{type_name}"
        if type_fullname == "qm.program._Program._Program":
            return self.visit(node._program)

        if hasattr(node, "DESCRIPTOR"):
            self.__visit(node)
        else:
            raise Exception("Failed to find descriptor on " + node)

    def __visit(self, node):
        if not hasattr(node, "DESCRIPTOR"):
            if type(node).__name__ in set(
                ["RepeatedCompositeFieldContainer", "RepeatedCompositeContainer"]
            ):
                for n in node[:]:
                    self.__visit(n)
            return
        descriptor = node.DESCRIPTOR
        if self.__call_enter(descriptor.full_name, node) is True:
            self.__call_visit(descriptor.full_name, node)
        self.__call_leave(descriptor.full_name, node)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qm/pb/errors.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='qm/pb/errors.proto',
  package='',
  syntax='proto3',
  serialized_options=b'\n\016qm.grpc.errorsB\013ErrorsProtoP\001\242\002\003HLW',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x12qm/pb/errors.proto*\x89\x02\n\x14JobManagerErrorTypes\x12\x1e\n\x1aJobManagerUnspecifiedError\x10\x00\x12\x14\n\x0fMissingJobError\x10\xe8\x07\x12#\n\x1eInvalidJobExecutionStatusError\x10\xe9\x07\x12(\n#InvalidOperationOnSimulatorJobError\x10\xea\x07\x12#\n\x1eInvalidOperationOnRealJobError\x10\xeb\x07\x12\x1e\n\x19JobOperationSpecificError\x10\xec\x07\x12\x15\n\x10\x43onfigQueryError\x10\xed\x07\"\x05\x08\x01\x10\xe7\x07\"\t\x08\xd0\x0f\x10\xff\xff\xff\xff\x07*\xb8\x02\n\x1eJobOperationSpecificErrorTypes\x12 \n\x1cJobOperationUnspecifiedError\x10\x00\x12\x1c\n\x17SingleInputElementError\x10\xb8\x17\x12!\n\x1cInvalidCorrectionMatrixError\x10\xb9\x17\x12-\n(ElementWithoutIntermediateFrequencyError\x10\xba\x17\x12&\n!InvalidDigitalInputThresholdError\x10\xbb\x17\x12#\n\x1eInvalidDigitalInputWindowError\x10\xbc\x17\x12%\n InvalidDigitalInputPolarityError\x10\xbd\x17\"\x05\x08\x01\x10\xb7\x17\"\t\x08\xa0\x1f\x10\xff\xff\xff\xff\x07*\xa0\x01\n\x15\x43onfigQueryErrorTypes\x12\x1f\n\x1b\x43onfigQueryUnspecifiedError\x10\x00\x12\x1b\n\x16MissingControllerError\x10\xa0\x1f\x12\x18\n\x13MissingElementError\x10\xa1\x1f\x12\x1d\n\x18MissingDigitalInputError\x10\xa2\x1f\"\x05\x08\x01\x10\x9f\x1f\"\t\x08\x88\'\x10\xff\xff\xff\xff\x07\x42%\n\x0eqm.grpc.errorsB\x0b\x45rrorsProtoP\x01\xa2\x02\x03HLWb\x06proto3'
)

_JOBMANAGERERRORTYPES = _descriptor.EnumDescriptor(
  name='JobManagerErrorTypes',
  full_name='JobManagerErrorTypes',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='JobManagerUnspecifiedError', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MissingJobError', index=1, number=1000,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='InvalidJobExecutionStatusError', index=2, number=1001,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='InvalidOperationOnSimulatorJobError', index=3, number=1002,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='InvalidOperationOnRealJobError', index=4, number=1003,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='JobOperationSpecificError', index=5, number=1004,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ConfigQueryError', index=6, number=1005,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=23,
  serialized_end=288,
)
_sym_db.RegisterEnumDescriptor(_JOBMANAGERERRORTYPES)

JobManagerErrorTypes = enum_type_wrapper.EnumTypeWrapper(_JOBMANAGERERRORTYPES)
_JOBOPERATIONSPECIFICERRORTYPES = _descriptor.EnumDescriptor(
  name='JobOperationSpecificErrorTypes',
  full_name='JobOperationSpecificErrorTypes',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='JobOperationUnspecifiedError', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SingleInputElementError', index=1, number=3000,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='InvalidCorrectionMatrixError', index=2, number=3001,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ElementWithoutIntermediateFrequencyError', index=3, number=3002,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='InvalidDigitalInputThresholdError', index=4, number=3003,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='InvalidDigitalInputWindowError', index=5, number=3004,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='InvalidDigitalInputPolarityError', index=6, number=3005,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=291,
  serialized_end=603,
)
_sym_db.RegisterEnumDescriptor(_JOBOPERATIONSPECIFICERRORTYPES)

JobOperationSpecificErrorTypes = enum_type_wrapper.EnumTypeWrapper(_JOBOPERATIONSPECIFICERRORTYPES)
_CONFIGQUERYERRORTYPES = _descriptor.EnumDescriptor(
  name='ConfigQueryErrorTypes',
  full_name='ConfigQueryErrorTypes',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ConfigQueryUnspecifiedError', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MissingControllerError', index=1, number=4000,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MissingElementError', index=2, number=4001,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MissingDigitalInputError', index=3, number=4002,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=606,
  serialized_end=766,
)
_sym_db.RegisterEnumDescriptor(_CONFIGQUERYERRORTYPES)

ConfigQueryErrorTypes = enum_type_wrapper.EnumTypeWrapper(_CONFIGQUERYERRORTYPES)
JobManagerUnspecifiedError = 0
MissingJobError = 1000
InvalidJobExecutionStatusError = 1001
InvalidOperationOnSimulatorJobError = 1002
InvalidOperationOnRealJobError = 1003
JobOperationSpecificError = 1004
ConfigQueryError = 1005
JobOperationUnspecifiedError = 0
SingleInputElementError = 3000
InvalidCorrectionMatrixError = 3001
ElementWithoutIntermediateFrequencyError = 3002
InvalidDigitalInputThresholdError = 3003
InvalidDigitalInputWindowError = 3004
InvalidDigitalInputPolarityError = 3005
ConfigQueryUnspecifiedError = 0
MissingControllerError = 4000
MissingElementError = 4001
MissingDigitalInputError = 4002


DESCRIPTOR.enum_types_by_name['JobManagerErrorTypes'] = _JOBMANAGERERRORTYPES
DESCRIPTOR.enum_types_by_name['JobOperationSpecificErrorTypes'] = _JOBOPERATIONSPECIFICERRORTYPES
DESCRIPTOR.enum_types_by_name['ConfigQueryErrorTypes'] = _CONFIGQUERYERRORTYPES
_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import random
import qm.program.expressions as _exp
from qm.qua._dsl import _Expression, _to_expression, declare, assign
from functools import wraps


def _library_function(lib_name, func_name):
    def library_decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            new_args = function(*args, **kwargs)
            return _Expression(
                _exp.lib_func(
                    lib_name, func_name, *[_to_expression(x) for x in new_args]
                )
            )

        return wrapper

    return library_decorator


class Math:
    @staticmethod
    @_library_function("math", "log")
    def log(x, base):
        """
        Computes :math:`\mathrm{log}_{base}(x)`

        :param x: a QUA fixed larger than pow2(-8)=0.00390625
        :param base: a QUA fixed larger than pow2(1/8)=1.09051
        :return: a QUA fixed
        """
        return (x, base)

    @staticmethod
    @_library_function("math", "pow")
    def pow(base, x):
        """
        Computes :math:`{base}^{x}`.
        Does not support base=1, nor the case where both base=0 & x=0.

        :param base: a non-negative QUA fixed
        :param x: a QUA fixed
        :return:
        """
        return (base, x)

    @staticmethod
    @_library_function("math", "div")
    def div(x, y):
        """
        Computes the division between two same-type variables :math:`x/y`

        :param x: a QUA parameter
        :param y: a QUA parameter not equal to 0
        :return: a QUA fixed
        """
        return (x, y)

    @staticmethod
    @_library_function("math", "exp")
    def exp(x):
        """
        Computes :math:`e^{x}`

        :param x: a QUA fixed smaller than ln(8)=2.0794415416
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "pow2")
    def pow2(x):
        """
        Computes :math:`2^{x}`

        :param x: a QUA fixed smaller than 3
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "ln")
    def ln(x):
        """
        Computes :math:`\mathrm{ln}(x)`

        :param x: a QUA fixed larger than exp(-8)=0.0003354627
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "log2")
    def log2(x):
        """
        Computes :math:`\mathrm{log}_{2}(x)`

        :param x: a QUA fixed larger than pow2(-8)=0.00390625
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "log10")
    def log10(x):
        """
        Computes :math:`\mathrm{log}_{10}(x)`

        :param x: a QUA fixed larger than pow10(-8)=0.00000001
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "sqrt")
    def sqrt(x):
        """
        Computes the square root of x

        :param x: a non-negative QUA fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "inv_sqrt")
    def inv_sqrt(x):
        """
        Computes the inverse square root of x

        :param x: a QUA fixed larger than 1/64
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "inv")
    def inv(x):
        """
        Computes the inverse of x

        :param x: a QUA fixed which is x<=-1/8 or 1/8<x
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "MSB")
    def msb(x):
        """
        Finds the index of the most significant bit in the parameter x.
        Notes:

        - Result is independent of sign, for example, +3 and -3 will return the same msb
        - The returned value will be the closet log2, rounded down.

          This is given by :math:`\mathrm{floor}(\mathrm{log}_2(|x|))`.

          For example:

          - msb(0.1) will return -4.
          - msb(5) will return 2.
        - For an integer, msb(0) will return 0.
        - For a fixed point number, msb(0) will return -28.

        :param x: a QUA fixed or a QUA int
        :return: a QUA int
        """
        return (x,)

    @staticmethod
    @_library_function("math", "elu")
    def elu(x):
        """
        Computes the Exponential Linear Unit activation function of x:
          :math:`\mathrm{ELU(x)} = \mathrm{max}(0, x) + \mathrm{min}(0, \mathrm{exp}(x)-1)`.

        :param x: a QUA fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "aelu")
    def aelu(x):
        """
        Computes faster an approximated Exponential Linear Unit activation function of x:
          :math:`\mathrm{aELU}(x) \sim \mathrm{ELU}(x)`

        :param x: a QUA fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "selu")
    def selu(x):
        """
        Computes the Scaled Exponential Linear Unit activation function of x:
          :math:`\mathrm{SELU}(x) = s*(\mathrm{max}(0, x)+a*\mathrm{min}(0, \mathrm{exp}(x)-1))`
            a=1.67326324, s=1.05070098

        :param x: a QUA fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "relu")
    def relu(x):
        """
        Computes the Rectified Linear Unit activation function of x:
          :math:`\mathrm{ReLU}(x) = \mathrm{max}(0, x)`

        :param x: a QUA fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "plrelu")
    def plrelu(x, a):
        """
        Computes the Parametric Leaky Rectified Linear Unit activation function of x:
          :math:`\mathrm{PLReLU}(x, a) = \mathrm{max}(0, x)+a*\mathrm{min}(0, x)`

        :param x: a QUA fixed
        :param a: a QUA fixed
        :return: a QUA fixed
        """
        return (x, a)

    @staticmethod
    @_library_function("math", "lrelu")
    def lrelu(x):
        """
        Computes the Leaky Rectified Linear Unit activation function of x:
          :math:`\mathrm{LReLU}(x)=\mathrm{max}(0, x)+0.01*\mathrm{min}(0, x)`

        :param x: a QUA fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "sin2pi")
    def sin2pi(x):
        """
        Computes :math:`\mathrm{sin}(2 \pi x)`.
        This is more efficient than Math.sin(2*np.pi*x).
        In addition, this function is immune to overflows: An overflow means that the argument gets a :math:`\pm 16`, which does not change the result due to the periodcity of the sine function.

        :param x: the angle in radians
        :type x: QUA variable of type fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "cos2pi")
    def cos2pi(x):
        """
        Computes :math:`\mathrm{cos}(2 \pi x)`.
        This is more efficient than Math.cos(:math:`2 \pi x`).
        In addition, this function is immune to overflows: An overflow means that the argument gets a :math:`\pm 16`, which does not change the result due to the periodcity of the cosine function.

        :param x: the angle in radians
        :type x: QUA variable of type fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "abs")
    def abs(x):
        """
        Computes the absolute value of x

        :param x: a QUA variable
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "sin")
    def sin(x):
        """
        Computes :math:`\mathrm{sin}(x)`

        :param x: the angle in radians
        :type x: QUA variable of type fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "cos")
    def cos(x):
        """
        Computes :math:`\mathrm{cos}(x)`

        :param x: the angle in radians
        :type x: QUA variable of type fixed
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "sum")
    def sum(x):
        """
        Computes the sum of an array x

        :param x: a QUA array
        :return: the sum of the array, has same type as x
        """
        return (x,)

    @staticmethod
    @_library_function("math", "max")
    def max(x):
        """
        Computes the max of an array x

        :param x: a QUA array
        :return: the max value of the array, has same type as x
        """
        return (x,)

    @staticmethod
    @_library_function("math", "min")
    def min(x):
        """
        Computes the min of an array x

        :param x: a QUA array
        :return: the min value of the array, has same type as x
        """
        return (x,)

    @staticmethod
    @_library_function("math", "argmax")
    def argmax(x):
        """
        Return the index of the maximum of an array

        :param x: a QUA array
        :return: the index of maximum value of array, a QUA Integer
        """
        return (x,)

    @staticmethod
    @_library_function("math", "argmin")
    def argmin(x):
        """
        Return the index of the minimum of an array

        :param x: a QUA array
        :return: the index of minimum value of array, a QUA Integer
        """
        return (x,)

    @staticmethod
    @_library_function("math", "dot")
    def dot(x, y):
        """
        Calculates a dot product of two QUA arrays of identical size.

        :param x: a QUA array
        :param y: a QUA array
        :return: The dot product of x and y, has same type as x and y

        :Example:
            >>> assign(c, dot(a, b))
        """
        return x, y


class Cast:
    @staticmethod
    @_library_function("cast", "mul_int_by_fixed")
    def mul_int_by_fixed(x, y):
        """
        Multiplies an int x by a fixed y, returning an int

        :param x: a QUA integer
        :param y: a QUA fixed
        :return: a QUA int which equals x*y
        """
        return x, y

    @staticmethod
    @_library_function("cast", "mul_fixed_by_int")
    def mul_fixed_by_int(x, y):
        """
        Multiplies a fixed x by an int y, returning a fixed

        :param x: a QUA fixed
        :param y: a QUA int
        :return: a QUA fixed which equals x*y
        """
        return x, y

    @staticmethod
    @_library_function("cast", "to_int")
    def to_int(x):
        """
        Casts a variable to int. Supports int, fixed or bool

        :param x: a QUA variable
        :return: a QUA int
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "to_fixed")
    def to_fixed(x):
        """
        Casts a variable to fixed. Supports int, fixed or bool

        :param x: a QUA variable
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "to_bool")
    def to_bool(x):
        """
        Casts a variable to bool. Supports int, fixed or bool

        :param x: a QUA variable
        :return: a QUA bool
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "unsafe_cast_int")
    def unsafe_cast_int(x):
        """
        Treats the given input variable, bitwise, as an integer.
        For a given fixed point number, this is equivalent to multiplying by
        :math:`2^{28}`

        Supports int, fixed or bool.

        :param x: a QUA variable
        :return: a QUA int
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "unsafe_cast_fixed")
    def unsafe_cast_fixed(x):
        """
        Treats the given input variable, bitwise, as a fixed point number.
        For a given integer, this is equivalent to multiplying by :math:`2^{-28}`

        Supports int, fixed or bool.

        :param x: a QUA variable
        :return: a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "unsafe_cast_bool")
    def unsafe_cast_bool(x):
        """
        Treats the given input variable, bitwise, as a boolean.
        A boolean is determined by the right-most bit, so for a given integer, this is
        equivalent to a parity check.

        Supports int, fixed or bool.

        :param x: a QUA variable
        :return: a QUA bool
        """
        return (x,)


class Util:
    @staticmethod
    @_library_function("util", "cond")
    def cond(condition, true_result, false_result):
        """
        Quick conditional operation. This is equivalent to a ternary operator available in some languges:
        i.e. a ? b : c, meaning 'b' if 'a' is true, or 'c' if 'a' is false.
        There is less computation overhead (less latency) when running this operation relative to the if conditional.

        :Example:
            >>> assign(var, cond(a, b, c)) #where a is a boolean expression

        """
        return condition, true_result, false_result


class Random:
    def __init__(self, seed=None):
        """
        A class for generating pseudo-random numbers in QUA

        :param seed: Optional. An integer seed for the pseudo-random number generator.
        """
        self._seed = declare(
            int, value=seed if seed is not None else random.randrange((1 << 28) - 1)
        )

    def set_seed(self, exp):
        """
        Set the seed for the pseudo-random number generator

        :param exp: a QUA expression
        """
        assign(self._seed, exp)

    @_library_function("random", "rand_int")
    def rand_int(self, max_int):
        """
        Returns a pseudorandom integer in range [0, max_int)

        :param max_int: maximum value

        :Example:
            >>> a= Random()
            >>> assign(b,a.rand_int(max_int))
        """
        return self._seed, max_int

    @_library_function("random", "rand_fixed")
    def rand_fixed(self):
        """
        Returns a pseudorandom fixed in range [0.0, 1.0)

        :Example:
            >>> a= Random()
            >>> assign(b,a.rand_fixed())
        """
        return (self._seed,)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from qm.pb import job_results_pb2 as qm_dot_pb_dot_job__results__pb2


class JobResultsServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetJobResultSchema = channel.unary_unary(
                '/qm.grpc.results_analyser.JobResultsService/GetJobResultSchema',
                request_serializer=qm_dot_pb_dot_job__results__pb2.GetJobResultSchemaRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobResultSchemaResponse.FromString,
                )
        self.GetJobState = channel.unary_unary(
                '/qm.grpc.results_analyser.JobResultsService/GetJobState',
                request_serializer=qm_dot_pb_dot_job__results__pb2.GetJobStateRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobStateResponse.FromString,
                )
        self.GetJobNamedResultHeader = channel.unary_unary(
                '/qm.grpc.results_analyser.JobResultsService/GetJobNamedResultHeader',
                request_serializer=qm_dot_pb_dot_job__results__pb2.GetJobNamedResultHeaderRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobNamedResultHeaderResponse.FromString,
                )
        self.GetJobNamedResult = channel.unary_stream(
                '/qm.grpc.results_analyser.JobResultsService/GetJobNamedResult',
                request_serializer=qm_dot_pb_dot_job__results__pb2.GetJobNamedResultRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobNamedResultResponse.FromString,
                )
        self.GetJobDebugData = channel.unary_stream(
                '/qm.grpc.results_analyser.JobResultsService/GetJobDebugData',
                request_serializer=qm_dot_pb_dot_job__results__pb2.GetJobDebugDataRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobDebugDataResponse.FromString,
                )
        self.GetJobErrors = channel.unary_unary(
                '/qm.grpc.results_analyser.JobResultsService/GetJobErrors',
                request_serializer=qm_dot_pb_dot_job__results__pb2.GetJobErrorsRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobErrorsResponse.FromString,
                )
        self.GetProgramMetadata = channel.unary_unary(
                '/qm.grpc.results_analyser.JobResultsService/GetProgramMetadata',
                request_serializer=qm_dot_pb_dot_job__results__pb2.GetProgramMetadataRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_job__results__pb2.GetProgramMetadataResponse.FromString,
                )


class JobResultsServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetJobResultSchema(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobState(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobNamedResultHeader(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobNamedResult(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobDebugData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobErrors(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetProgramMetadata(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_JobResultsServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetJobResultSchema': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobResultSchema,
                    request_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobResultSchemaRequest.FromString,
                    response_serializer=qm_dot_pb_dot_job__results__pb2.GetJobResultSchemaResponse.SerializeToString,
            ),
            'GetJobState': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobState,
                    request_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobStateRequest.FromString,
                    response_serializer=qm_dot_pb_dot_job__results__pb2.GetJobStateResponse.SerializeToString,
            ),
            'GetJobNamedResultHeader': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobNamedResultHeader,
                    request_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobNamedResultHeaderRequest.FromString,
                    response_serializer=qm_dot_pb_dot_job__results__pb2.GetJobNamedResultHeaderResponse.SerializeToString,
            ),
            'GetJobNamedResult': grpc.unary_stream_rpc_method_handler(
                    servicer.GetJobNamedResult,
                    request_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobNamedResultRequest.FromString,
                    response_serializer=qm_dot_pb_dot_job__results__pb2.GetJobNamedResultResponse.SerializeToString,
            ),
            'GetJobDebugData': grpc.unary_stream_rpc_method_handler(
                    servicer.GetJobDebugData,
                    request_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobDebugDataRequest.FromString,
                    response_serializer=qm_dot_pb_dot_job__results__pb2.GetJobDebugDataResponse.SerializeToString,
            ),
            'GetJobErrors': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobErrors,
                    request_deserializer=qm_dot_pb_dot_job__results__pb2.GetJobErrorsRequest.FromString,
                    response_serializer=qm_dot_pb_dot_job__results__pb2.GetJobErrorsResponse.SerializeToString,
            ),
            'GetProgramMetadata': grpc.unary_unary_rpc_method_handler(
                    servicer.GetProgramMetadata,
                    request_deserializer=qm_dot_pb_dot_job__results__pb2.GetProgramMetadataRequest.FromString,
                    response_serializer=qm_dot_pb_dot_job__results__pb2.GetProgramMetadataResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qm.grpc.results_analyser.JobResultsService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class JobResultsService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetJobResultSchema(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.results_analyser.JobResultsService/GetJobResultSchema',
            qm_dot_pb_dot_job__results__pb2.GetJobResultSchemaRequest.SerializeToString,
            qm_dot_pb_dot_job__results__pb2.GetJobResultSchemaResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.results_analyser.JobResultsService/GetJobState',
            qm_dot_pb_dot_job__results__pb2.GetJobStateRequest.SerializeToString,
            qm_dot_pb_dot_job__results__pb2.GetJobStateResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobNamedResultHeader(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.results_analyser.JobResultsService/GetJobNamedResultHeader',
            qm_dot_pb_dot_job__results__pb2.GetJobNamedResultHeaderRequest.SerializeToString,
            qm_dot_pb_dot_job__results__pb2.GetJobNamedResultHeaderResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobNamedResult(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/qm.grpc.results_analyser.JobResultsService/GetJobNamedResult',
            qm_dot_pb_dot_job__results__pb2.GetJobNamedResultRequest.SerializeToString,
            qm_dot_pb_dot_job__results__pb2.GetJobNamedResultResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobDebugData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/qm.grpc.results_analyser.JobResultsService/GetJobDebugData',
            qm_dot_pb_dot_job__results__pb2.GetJobDebugDataRequest.SerializeToString,
            qm_dot_pb_dot_job__results__pb2.GetJobDebugDataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobErrors(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.results_analyser.JobResultsService/GetJobErrors',
            qm_dot_pb_dot_job__results__pb2.GetJobErrorsRequest.SerializeToString,
            qm_dot_pb_dot_job__results__pb2.GetJobErrorsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetProgramMetadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.results_analyser.JobResultsService/GetProgramMetadata',
            qm_dot_pb_dot_job__results__pb2.GetProgramMetadataRequest.SerializeToString,
            qm_dot_pb_dot_job__results__pb2.GetProgramMetadataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from qm.pb.frontend_pb2 import SimulationRequest
from qm.simulate.interface import SimulatorInterface


class RawInterface(SimulatorInterface):
    """
    Creates a raw interface for use in :class:`~qm.simulate.interface.SimulationConfig`.
    A raw interface defines samples that will be inputted into the OPX inputs.

    :param list connections:

        List of tuples with the connection. Each tuple should be:

                ``(toController: str, toPort: int, toSamples: List[float])``
    :param float noisePower: How much noise to add to the input.

    Example::

    >>> job = qm.simulate(prog2, SimulationConfig(
    >>>                   duration=20000,
    >>>                   # 500 ns of DC 0.2 V into con1 input 1
    >>>                   simulation_interface=RawInterface([("con1", 1, [0.2]*500)])
    """

    def __init__(self, connections, noisePower=0.0):

        if (
            type(noisePower) is not float and type(noisePower) is not int
        ) or noisePower < 0:
            raise Exception("noisePower must be a positive number")

        self.noisePower = noisePower

        if type(connections) is not list:
            raise Exception("connections argument must be of type list")

        self.connections = list()
        for connection in connections:
            if type(connection) is not tuple:
                raise Exception("each connection must be of type tuple")
            if len(connection) == 3:
                if (
                    type(connection[0]) is not str
                    or type(connection[1]) is not int
                    or type(connection[2]) is not list
                ):
                    raise Exception(
                        "connection should be (fromController, fromPort, toSamples)"
                    )
                self.connections.append(connection)
            else:
                raise Exception("connection should be tuple of length 3 or 4")

    def update_simulate_request(self, request: SimulationRequest):
        request.simulate.simulationInterface.raw.SetInParent()
        request.simulate.simulationInterface.raw.noisePower = self.noisePower
        for connection in self.connections:
            con = request.simulate.simulationInterface.raw.connections.add()
            con.fromController = connection[0]
            con.fromPort = connection[1]
            con.toSamples.extend(connection[2])



from dataclasses import dataclass
from enum import Enum
from typing import List

from qm.pb.job_results_pb2 import GetJobErrorsRequest, GetJobErrorsResponse
from qm.pb.job_results_pb2_grpc import JobResultsServiceStub


class ExecutionErrorSeverity(Enum):
    Warn = 0
    Error = 1


@dataclass(frozen=True)
class ExecutionError:
    error_code: int
    message: str
    severity: ExecutionErrorSeverity

    def __repr__(self):
        return f"{self.error_code}\t\t{self.severity.name}\t\t{self.message}"


class ExecutionReport:
    def __init__(self, job_id: str, service: JobResultsServiceStub) -> None:
        super().__init__()
        self._errors: List[ExecutionError] = []
        self._job_id: str = job_id
        self._service: JobResultsServiceStub = service
        self._errors = ExecutionReport._load_errors(self._job_id, self._service)

    @staticmethod
    def _load_errors(
        job_id: str, service: JobResultsServiceStub
    ) -> List[ExecutionError]:
        request = GetJobErrorsRequest()
        request.jobId = job_id
        response: GetJobErrorsResponse = service.GetJobErrors(request)
        return [
            ExecutionError(
                error_code=item.errorCode,
                message=item.message,
                severity=ExecutionReport._parse_error_severity(item.errorSeverity),
            )
            for item in response.errors
        ]

    @staticmethod
    def _parse_error_severity(error_severity) -> ExecutionErrorSeverity:
        if error_severity == GetJobErrorsResponse.WARNING:
            return ExecutionErrorSeverity.Warn
        elif error_severity == GetJobErrorsResponse.ERROR:
            return ExecutionErrorSeverity.Error

    def has_errors(self) -> bool:
        """
        :return: True if encountered a runtime error while executing the job.
        """
        return len(self._errors) > 0

    def errors(self) -> List[ExecutionError]:
        """

        :return: list of all execution errors for this job
        """
        return self._errors.copy()

    def _report_header(self) -> str:
        return (
            f"Execution report for job {self._job_id}\nErrors:\n"
            f"Please refer to section: Error Indications and Error Reporting in documentation for additional information\n\n"
            "code\t\tseverity\tmessage"
        )

    def __repr__(self) -> str:
        if self.has_errors():
            errorsStr = ""
            for error in self._errors:
                errorsStr += "\n" + str(error)
            return self._report_header() + errorsStr
        else:
            return f"Execution report for job {self._job_id}\nNo errors"



import numpy as np


class SimulatorSamples(object):
    def __init__(self, controllers):
        for k, v in controllers.items():
            self.__setattr__(k, v)

    @staticmethod
    def from_np_array(arr: np.ndarray):
        controllers = dict()
        for col in arr.dtype.names:
            parts = col.split(":")
            controller = controllers.setdefault(
                parts[0], {"analog": dict(), "digital": dict()}
            )
            controller[parts[1]][parts[2]] = arr[col]
        res = dict()
        for item in controllers.items():
            res[item[0]] = SimulatorControllerSamples(
                item[1]["analog"], item[1]["digital"]
            )
        return SimulatorSamples(res)


class SimulatorControllerSamples(object):
    def __init__(self, analog, digital):
        self.analog = analog
        self.digital = digital

    def plot(self, analog_ports=None, digital_ports=None):
        """
        Plots the simulated output of the OPX in the given ports.
        If no ports are given, all active ports are plotted.

        :param analog_ports: Union[None, str, list[str]]
        :param digital_ports: Union[None, str, list[str]]
        """
        import matplotlib.pyplot as plt

        for port, samples in self.analog.items():
            if analog_ports is None or port in analog_ports:
                plt.plot(samples, label=f"Analog {port}")
        for port, samples in self.digital.items():
            if digital_ports is None or port in digital_ports:
                plt.plot(samples, label=f"Digital {port}")
        plt.xlabel("Time [ns]")
        plt.ylabel("Output")
        plt.legend()



from qm.simulate.interface import (  # noqa
    SimulationConfig,
    InterOpxAddress,
    InterOpxChannel,
    ControllerConnection,
    InterOpxPairing,
)
from qm.simulate.loopback import LoopbackInterface  # noqa



import json as _json
from dataclasses import dataclass
from enum import Enum
from io import BytesIO
from typing import Tuple

import numpy
import numpy.lib.format as _format
from google.protobuf.json_format import MessageToDict

from qm._QmJobErrors import (
    _handle_job_manager_error,
    MissingElementError,
    ElementWithoutIntermediateFrequencyError,
    ElementWithSingleInputError,
    InvalidElementCorrectionError,
    MissingDigitalInputError,
    InvalidDigitalInputThresholdError,
    InvalidDigitalInputWindowError,
    InvalidDigitalInputPolarityError,
)
from qm._report import ExecutionReport
from qm._results import JobResults
from qm.logger import logger
from qm.pb.frontend_pb2 import (
    IsJobRunningRequest,
    IsJobAcquiringDataRequest,
    IsJobAcquiringDataResponse,
    HaltRequest,
    GetSimulatedQuantumStateRequest,
)
from qm.pb.frontend_pb2 import ResumeRequest, PausedStatusRequest
from qm.pb.job_manager_pb2 import (
    SetElementCorrectionRequest,
    GetElementCorrectionRequest,
    SetDigitalInputThresholdRequest,
    GetDigitalInputThresholdRequest,
    SetDigitalInputWindowRequest,
    GetDigitalInputWindowRequest,
    SetDigitalInputPolarityRequest,
    ACTIVE_HIGH,
    ACTIVE_LOW,
    GetDigitalInputPolarityRequest,
)
from qm.pb.job_manager_pb2_grpc import JobManagerServiceStub
from qm.pb.job_results_pb2 import PullSimulatorSamplesRequest
from qm.pb.job_results_pb2_grpc import JobResultsServiceStub
from qm.persistence import BinaryAsset
from qm.results.SimulatorSamples import SimulatorSamples


class AcquiringStatus(Enum):
    AcquireStopped = 0
    NoDataToAcquire = 1
    HasDataToAcquire = 2


@dataclass
class DensityMatrix:
    timestamp: int
    data: numpy.ndarray


class SimulatorOutput:
    def __init__(self, job_id: str, frontend) -> None:
        super().__init__()
        self._id = job_id
        self._frontend = frontend

    def get_quantum_state(self) -> DensityMatrix:
        request = GetSimulatedQuantumStateRequest()
        request.jobId = self._id

        result = self._frontend.GetSimulatedQuantumState(request)
        if result.ok:
            flatten = numpy.array(
                [complex(item.re, item.im) for item in result.state.data]
            )
            n = int(numpy.sqrt(len(flatten)))
            if len(flatten) != (n * n):
                raise RuntimeError("Quantum state matrix is not correct")
            else:
                pass

            matrix = flatten.reshape(n, n)
            timestamp = result.state.timeStamp
            return DensityMatrix(timestamp, matrix)
        else:
            raise RuntimeError("Error while pulling quantum state")


class QmJob:
    def __init__(
        self,
        qmm: "qm.QuantumMachinesManager.QuantumMachinesManager",
        job_id: str,
        execute_response=None,
    ):
        super().__init__()
        self._id = job_id
        self._logger = logger.getChild("job")

        self._simulated_analog_outputs = {"samples": None, "waveforms": None}
        self._simulated_digital_outputs = {"samples": None, "waveforms": None}
        if execute_response is not None and execute_response.HasField("simulated"):
            if execute_response.simulated.HasField("analogOutputs"):
                self._simulated_analog_outputs = MessageToDict(
                    execute_response.simulated.analogOutputs
                )
            if execute_response.simulated.HasField("digitalOutputs"):
                self._simulated_digital_outputs = MessageToDict(
                    execute_response.simulated.digitalOutputs
                )
            if len(execute_response.simulated.errors) > 0:
                raise RuntimeError("\n".join(execute_response.simulated.errors))

        self._qmm = qmm
        self._frontend = qmm._frontend
        self._job_manager_client = JobManagerServiceStub(self.manager._channel)

    def id(self):
        """
        :return: The id of the job
        """
        return self._id

    @property
    def manager(self) -> "qm.QuantumMachinesManager.QuantumMachinesManager":
        """
        The QM object where this job lives

        :return:

        """
        return self._qmm

    def simulated_analog_waveforms(self):
        """
        Return the results of the simulation of quantum elements and analog outputs.

        The returned dictionary has the following keys and entries:

        ``elements``: a dictionary containing the outputs with timestamps and values arranged by quantum elements.

        ``controllers``: a dictionary containing the outputs with timestamps and values arranged by controllers.

            ``ports``: a dictionary containing the outputs with timestamps and values arranged by output ports.

                for each element or output port, the entry is a list of dictionaries with the following information:

                ``timestamp``:

                    The time, in nsec, from the start of the program to the start of the pulse.

                ``samples``:

                    Output information, with ``duration`` given in nsec and ``value`` given normalized OPX output units.


        :return: A dictionary containing output information for the analog outputs of the controller.

        """
        return self._simulated_analog_outputs["waveforms"]

    def simulated_digital_waveforms(self):
        """
        Return the results of the simulation of digital outputs.

        ``controllers``: a dictionary containing the outputs with timestamps and values arranged by controllers.

            ``ports``: a dictionary containing the outputs with timestamps and values arranged by output ports.

                for each element or output port, the entry is a list of dictionaries with the following information:

                ``timestamp``:

                    The time, in nsec, from the start of the program to the start of the pulse.

                ``samples``:

                    A list containing the sequence of outputted values, with ``duration`` given in nsec
                    and ``value`` given as a boolean value.



        :return: A dictionary containing output information for the analog outputs of the controller.
        """
        return self._simulated_digital_outputs["waveforms"]

    def halt(self) -> bool:
        """
        Halts the job on the opx
        """
        request = HaltRequest()
        request.jobId = self._id

        response = self._frontend.Halt(request)
        return response.ok

    def _get_np_simulated_samples(
        self, path=None, include_analog=True, include_digital=True
    ):
        request = PullSimulatorSamplesRequest()
        request.jobId = self._id
        request.includeAnalog = include_analog
        request.includeDigital = include_digital

        writer = BytesIO()
        data_writer = BytesIO()

        for result in self._frontend.PullSimulatorSamples(request):
            if result.ok:
                one_of = result.WhichOneof("body")
                if one_of == "header":
                    _format.write_array_header_2_0(
                        writer,
                        {
                            "descr": _json.loads(result.header.simpleDType),
                            "fortran_order": False,
                            "shape": (result.header.countOfItems,),
                        },
                    )
                elif one_of == "data":
                    data_writer.write(result.data.data)

            else:
                raise RuntimeError("Error while pulling samples")

        data_writer.seek(0)
        for d in data_writer:
            writer.write(d)

        writer.seek(0)
        return numpy.load(writer)

    def get_simulated_samples(self, include_analog=True, include_digital=True):
        """
        Obtain the output samples of a QUA program simulation.

        Samples are returned in an object that holds the controllers in the current simulation,
        where each controller's name will be a property of this object.
        The value of each property of the returned value is an object with the following properties:

        ``analog``:

            holds a dictionary with analog port names as keys and numpy array of samples as values.

        ``digital``:

            holds a dictionary with digital port names as keys and numpy array of samples as values.

        It is also possible to directly plot the outputs using a built-in plot command.

        Example::

            >>> samples = job.get_simulated_samples()
            >>> analog1 = samples.con1.analog["1"]  # obtain analog port 1 of controller "con1"
            >>> digital9 = samples.con1.analog["9"]  # obtain digital port 9 of controller "con1"
            >>> samples.con1.plot()  # Plots all active ports
            >>> samples.con1.plot(analog_ports=['1', '2'], digital_ports=['9'])  # Plots the given output ports

        .. note::

            The simulated digital waveforms are delayed by 136ns relative to the real
            output of the OPX.

        :param include_analog: Should we collect simulated analog samples
        :param include_digital: Should we collect simulated digital samples

        :return: The simulated samples of the job.
        """
        return SimulatorSamples.from_np_array(
            self._get_np_simulated_samples(
                include_analog=include_analog, include_digital=include_digital
            )
        )

    def resume(self):
        """
        Resumes a program that was halted using the pause statement
        """
        request = ResumeRequest()
        request.jobId = self._id
        self._frontend.Resume(request)

    @property
    def result_handles(self) -> JobResults:
        """
        :type: qm._results.JobResults
        :return: The handles that this job generated
        """
        has_job_streaming_state = self.manager._caps.has_job_streaming_state
        if not has_job_streaming_state:
            import qm._results_01

            return qm._results_01.JobResults(
                self._id,
                JobResultsServiceStub(self.manager._channel),
                self.manager.store,
            )
        return JobResults(
            self._id, JobResultsServiceStub(self.manager._channel), self.manager.store
        )

    @property
    def simulator(self) -> SimulatorOutput:
        return SimulatorOutput(self._id, self._frontend)

    def set_element_correction(
        self, element, correction
    ) -> Tuple[float, float, float, float]:
        """
        Sets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with a quantum element.

        Changes will only be done to the current job!

        Values will be rounded to an accuracy of 2^-16.
        Valid values for the correction values are between -2 and (2 - 2^-16).

        Warning - the correction matrix can increase the output voltage which might result in an
        overflow.

        :param element: the name of the element to update the correction for
        :type element: str
        :param correction:

            tuple is of the form (v00, v01, v10, v11) where
            the matrix is
            | v00 v01 |
            | v10 v11 |

        :type correction: tuple
        :return: The correction matrix, after rounding to the OPX resolution.
        """
        request = SetElementCorrectionRequest()
        request.jobId = self._id
        request.qeName = element
        request.correction.v00 = correction[0]
        request.correction.v01 = correction[1]
        request.correction.v10 = correction[2]
        request.correction.v11 = correction[3]

        response = self._job_manager_client.SetElementCorrection(request)
        valid_errors = (
            MissingElementError,
            ElementWithSingleInputError,
            InvalidElementCorrectionError,
            ElementWithoutIntermediateFrequencyError,
        )
        _handle_job_manager_error(request, response, valid_errors)
        return (
            response.correction.v00,
            response.correction.v01,
            response.correction.v10,
            response.correction.v11,
        )

    def get_element_correction(self, element) -> Tuple[float, float, float, float]:
        """
        Gets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with a quantum element.

        :param element: the name of the element to update the correction for
        :type element: str
        :return: The current correction matrix
        """
        request = GetElementCorrectionRequest()
        request.jobId = self._id
        request.qeName = element

        response = self._job_manager_client.GetElementCorrection(request)
        valid_errors = (
            MissingElementError,
            ElementWithSingleInputError,
            ElementWithoutIntermediateFrequencyError,
        )
        _handle_job_manager_error(request, response, valid_errors)
        return (
            response.correction.v00,
            response.correction.v01,
            response.correction.v10,
            response.correction.v11,
        )

    def is_paused(self):
        """
        :return: Returns ``True`` if the job is in a paused state.

        see also:

            ``resume()``
        """
        request = PausedStatusRequest()
        request.jobId = self._id

        response = self._frontend.PausedStatus(request)
        return response.isPaused

    def _is_job_running(self):
        """
        :return: Returns ``True`` if the job is running
        """
        request = IsJobRunningRequest()
        request.jobId = self._id

        response = self._frontend.IsJobRunning(request)
        return response.isRunning

    def _is_data_acquiring(self):
        """
        Returns the data acquiring status.
        The possible statuses are: AcquireStopped, NoDataToAcquire,  HasDataToAcquire

        :return: An AcquiringStatus enum object
        """
        request = IsJobAcquiringDataRequest()
        request.jobId = self._id

        response = self._frontend.IsJobAcquiringData(request)
        result = response.acquiringStatus
        if result == IsJobAcquiringDataResponse.ACQUIRE_STOPPED:
            return AcquiringStatus(0)
        elif result == IsJobAcquiringDataResponse.NO_DATA_TO_ACQUIRE:
            return AcquiringStatus(1)
        elif result == IsJobAcquiringDataResponse.HAS_DATA_TO_ACQUIRE:
            return AcquiringStatus(2)

    def execution_report(self) -> ExecutionReport:
        """
        Get runtime errors report for this job. See :ref:`Guides/error:Runtime Errors`.

        :type: qm._report.ExecutionReport
        :return: An object holding the errors that this job generated.
        """
        return ExecutionReport(self._id, JobResultsServiceStub(self.manager._channel))

    def set_digital_input_threshold(
        self, port: Tuple[str, int], threshold: float
    ) -> float:
        request = SetDigitalInputThresholdRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]
        request.threshold = threshold

        response = self._job_manager_client.SetDigitalInputThreshold(request)
        valid_errors = (MissingDigitalInputError, InvalidDigitalInputThresholdError)
        _handle_job_manager_error(request, response, valid_errors)
        return response.threshold

    def get_digital_input_threshold(self, port: Tuple[str, int]) -> float:
        request = GetDigitalInputThresholdRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]

        response = self._job_manager_client.GetDigitalInputThreshold(request)
        valid_errors = (MissingDigitalInputError,)
        _handle_job_manager_error(request, response, valid_errors)
        return response.threshold

    def set_digital_input_window(self, port: Tuple[str, int], window: int) -> int:
        request = SetDigitalInputWindowRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]
        request.window = window

        response = self._job_manager_client.SetDigitalInputWindow(request)
        valid_errors = (MissingDigitalInputError, InvalidDigitalInputWindowError)
        _handle_job_manager_error(request, response, valid_errors)
        return response.window

    def get_digital_input_window(self, port: Tuple[str, int]) -> int:
        request = GetDigitalInputWindowRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]

        response = self._job_manager_client.GetDigitalInputWindow(request)
        valid_errors = (MissingDigitalInputError,)
        _handle_job_manager_error(request, response, valid_errors)
        return response.window

    def set_digital_input_polarity(self, port: Tuple[str, int], polarity: str) -> str:
        request = SetDigitalInputPolarityRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]
        request.polarity = ACTIVE_HIGH
        if polarity == "HIGH":
            request.polarity = ACTIVE_HIGH
        elif polarity == "LOW":
            request.polarity = ACTIVE_LOW
        else:
            raise InvalidDigitalInputPolarityError(
                f"Invalid value for polarity {polarity}. Valid values are: 'HIGH' "
                f"or 'LOW'"
            )

        response = self._job_manager_client.SetDigitalInputPolarity(request)
        valid_errors = (MissingDigitalInputError, InvalidDigitalInputPolarityError)
        _handle_job_manager_error(request, response, valid_errors)
        return response.polarity

    def get_digital_input_polarity(self, port: Tuple[str, int]) -> str:
        request = GetDigitalInputPolarityRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]

        response = self._job_manager_client.GetDigitalInputPolarity(request)
        valid_errors = (MissingDigitalInputError,)
        _handle_job_manager_error(request, response, valid_errors)
        if response.polarity == ACTIVE_HIGH:
            return "HIGH"
        elif response.polarity == ACTIVE_LOW:
            return "LOW"
        else:
            raise Exception(
                f"Received invalid digital input polarity value {response.polarity} from server."
            )


class _SavedResults:
    def __init__(self, asset: BinaryAsset) -> None:
        self._asset = asset
        super().__init__()

    def get_numpy_results(self, *args, **kwargs):
        input = self._asset.for_reading()
        try:
            return numpy.load(input)
        finally:
            input.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qm',
 'qm.grpc',
 'qm.grpc.quantum_simulator',
 'qm.grpc_client_interceptor',
 'qm.pb',
 'qm.program',
 'qm.qua',
 'qm.qua_future',
 'qm.results',
 'qm.simulate']

package_data = \
{'': ['*']}

install_requires = \
['betterproto>=1.2.5,<2.0.0',
 'grpcio>=1.39.0,<2.0.0',
 'marshmallow-polyfield>=5.7,<6.0',
 'marshmallow>=3.0.0,<4.0.0',
 'numpy>=1.17.0,<2.0.0',
 'protobuf>=3.7.1,<4.0.0',
 'qua>=0.1.0,<0.2.0']

setup_kwargs = {
    'name': 'qm-qua',
    'version': '0.3.3',
    'description': 'QUA language SDK to control a Quantum Computer',
    'long_description': '\n# QUA SDK\n\nQUA language SDK to control a Quantum Computer',
    'author': 'Tal Shani',
    'author_email': 'tal@quantum-machines.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from qm._deprecated import deprecated
from qm.qua import Math


@deprecated("use Math.sin2pi instead", "0.6", "0.8")
def sin2pi(x):
    """
    Compute the sin of a 2*pi*x on hardware
    :param x: the angle in radians
    :return:
    """
    return Math.sin2pi(x)


@deprecated("use Math.cos2pi instead", "0.6", "0.8")
def cos2pi(x):
    """
    Compute the cos of a 2*pi*x on hardware
    :param x: the angle in radians
    :return:
    """
    return Math.cos2pi(x)


@deprecated("use Math.abs instead", "0.6", "0.8")
def abs(x):
    """
    Compute the absolute value of x on hardware
    :param x: a QUA variable
    :return:
    """
    return Math.abs(x)


@deprecated("use Math.sin instead", "0.6", "0.8")
def sin(x):
    """
    Compute the sin of a x on hardware
    :param x: the angle in radians
    :return:
    """
    return Math.sin(x)


@deprecated("use Math.cos instead", "0.6", "0.8")
def cos(x):
    """
    Compute the cos of a x on hardware
    :param x: the angle in radians
    :return:
    """
    return Math.cos(x)


@deprecated("use Math.sum instead", "0.6", "0.8")
def sum(x):
    """
    Compute sum of an array x on hardware
    :param x: a QUA array
    :return: sum of the array, has same type as x
    """
    return Math.sum(x)


@deprecated("use Math.max instead", "0.6", "0.8")
def max(x):
    """
    Compute max of an array x on hardware
    :param x: a QUA array
    :return: max value of the array, has same type as x
    """
    return Math.max(x)


@deprecated("use Math.min instead", "0.6", "0.8")
def min(x):
    """
    Compute min of an array x on hardware
    :param x: a QUA array
    :return: min value of the array, has same type as x
    """
    return Math.min(x)


@deprecated("use Math.argmax instead", "0.6", "0.8")
def argmax(x):
    """
    Return the index of the maximum of an array on hardware
    :param x: a QUA array
    :return: index of maximum value of array
    """
    return Math.argmax(x)


@deprecated("use Math.argmin instead", "0.6", "0.8")
def argmin(x):
    """
    Return the index of the minimum of an array on hardware
    :param x: a QUA array
    :return: index of minimum value of array
    """
    return Math.argmin(x)
